#pragma ident "$Id: send.c,v 1.7 2004/01/29 18:45:50 dechavez Exp $"
/*======================================================================
 * 
 * Send a frame to the peer.
 *
 *====================================================================*/
#include "iacp.h"
#include "util.h"

/* Set various values in the handle's send field */

static VOID SetSendError(IACP *iacp, int value)
{
    MUTEX_LOCK(&iacp->mutex);
        iacp->send.error = value;
    MUTEX_UNLOCK(&iacp->mutex);
}

static VOID UpdateSendTstamp(IACP *iacp)
{
time_t now;

    now = time(NULL);

    MUTEX_LOCK(&iacp->mutex);
        iacp->send.tstamp = now;
    MUTEX_UNLOCK(&iacp->mutex);
}

static VOID IncrSendNbyte(IACP *iacp, int increment)
{
    MUTEX_LOCK(&iacp->mutex);
        iacp->send.nbyte += increment;
    MUTEX_UNLOCK(&iacp->mutex);
}

static VOID IncrSendNframe(IACP *iacp)
{
    MUTEX_LOCK(&iacp->mutex);
        ++iacp->send.nframe;
    MUTEX_UNLOCK(&iacp->mutex);
}

static UINT32 NextSeqno(IACP *iacp)
{
UINT32 retval;

    MUTEX_LOCK(&iacp->mutex);
        retval = iacp->seqno++;
    MUTEX_UNLOCK(&iacp->mutex);
    return retval;
}

/* Some errors should be ignored */

static BOOL IgnorableError(int error)
{
    switch (error) {
      case EWOULDBLOCK:
#if (EWOULDBLOCK != EAGAIN)
      case EAGAIN:
#endif
      case EINTR:
        return TRUE;
    }

    return FALSE;
}

/* Local function for doing timeout enabled writes.  This assumes that
 * the file descripter has already been set for non-blocking I/O,
 * which is done as part of iacpOpen() for the clients, and iacpAccept()
 * for the servers.
 */

static BOOL write_to(IACP *iacp, UINT8 *buf, UINT32 want)
{
size_t remain;
ssize_t WriteResult;
UINT8 *ptr;
int SelectResult;
int LogLevel;
UINT32 timeoutUS;
#define USEC_PER_MSEC 1000
#define USEC_PER_SEC  1000000
fd_set writefds;
struct timeval timeout;
static char *fid = "write_to";

    LogLevel = iacpGetDebug(iacp) ? LOG_ERR : LOG_DEBUG;

    if (want == 0) return TRUE;

    if (buf == NULL) {
        errno = EINVAL;
        SetSendError(iacp, errno);
        logioMsg(iacp->lp, LogLevel, "%s: NULL buf!", iacpPeerIdent(iacp), fid);
        return FALSE;
    }

    SelectResult = 0;
    timeoutUS = iacpGetTimeoutInterval(iacp) * USEC_PER_MSEC;
    timeout.tv_sec  = timeoutUS / USEC_PER_SEC;
    timeout.tv_usec = timeoutUS - (timeout.tv_sec * USEC_PER_SEC);

/* Mask out our file descriptor as the only one to look at */

    FD_ZERO(&writefds);
    FD_SET(iacp->sd, &writefds);
    
/*  Write to socket until desired number of bytes sent */

    remain = want;
    ptr    = buf;

    while (remain) {
        SelectResult = select(FD_SETSIZE, NULL, &writefds, NULL, &timeout);
        SetSendError(iacp, errno);
        if (SelectResult == 0) {
            errno = ETIMEDOUT;
            SetSendError(iacp, errno);
            logioMsg(iacp->lp, LogLevel, "%s: %s: select: %s", iacpPeerIdent(iacp), fid, strerror(errno));
            return FALSE;
        } else if (SelectResult < 0) {
            if (!IgnorableError(errno)) {
                logioMsg(iacp->lp, LogLevel, "%s: %s: select: %s", iacpPeerIdent(iacp), fid, strerror(errno));
                return FALSE;
            } else {
                logioMsg(iacp->lp, LOG_ERR, "%s: %s: select: %s (IGNORED)", iacpPeerIdent(iacp), fid, strerror(errno));
            }
        } else {
            WriteResult = write(iacp->sd, (void *) ptr, remain);
            SetSendError(iacp, errno);
            if (WriteResult > 0) {
                remain -= WriteResult;
                ptr    += WriteResult;
            } else if (WriteResult == 0) {
                errno = ECONNRESET;
                SetSendError(iacp, errno);
                logioMsg(iacp->lp, LogLevel, "%s: %s: write: %s", iacpPeerIdent(iacp), fid, strerror(errno));
                return FALSE;
            } else if (!IgnorableError(errno)) {
                logioMsg(iacp->lp, LogLevel, "%s: %s: write: %s", iacpPeerIdent(iacp), fid, strerror(errno));
                return FALSE;
            } else {
                logioMsg(iacp->lp, LOG_ERR, "%s: %s: write: %s (IGNORED)", iacpPeerIdent(iacp), fid, strerror(errno));
            }
        }
    }

    UpdateSendTstamp(iacp);
    SetSendError(iacp, 0);
    IncrSendNbyte(iacp, want);
    return TRUE;
}

/* Send an IACP frame */

BOOL iacpSendFrame(IACP *iacp, IACP_FRAME *frame)
{
UINT8 *ptr;
UINT8 tmpbuf[IACP_MAX_PREAMBLE_LEN];
static char *fid = "iacpSendFrame";

    if (iacp == (IACP *) NULL || frame == (IACP_FRAME *) NULL) {
        errno = EINVAL;
        return FALSE;
    }

    /* build and send payload preamble */

    frame->seqno = NextSeqno(iacp);

    ptr = tmpbuf;
    ptr += utilPackBytes(ptr, (UINT8 *) IACP_TAG, IACP_TAGLEN);
    ptr += utilPackUINT32(ptr, frame->seqno);
    ptr += utilPackUINT32(ptr, frame->payload.type);
    ptr += utilPackUINT32(ptr, frame->payload.len);
    if (!write_to(iacp, tmpbuf, IACP_DATA_PREAMBLE_LEN)) return FALSE;

    /* send payload */

    if (!write_to(iacp, frame->payload.data, frame->payload.len)) return FALSE;

    /* sign frame and build and send signature preamble */

    iacpSignFrame(iacp, frame);
    ptr = tmpbuf;
    ptr += utilPackUINT32(ptr, frame->auth.id);
    ptr += utilPackUINT32(ptr, frame->auth.len);
    if (!write_to(iacp, tmpbuf, IACP_SIG_PREAMBLE_LEN)) return FALSE;

    /* send the signature */

    if (!write_to(iacp,  frame->auth.sig, frame->auth.len)) return FALSE;

    IncrSendNframe(iacp);

    if (iacpGetDebug(iacp) > 1) {
        logioMsg(iacp->lp, LOG_INFO, "%s: %s: frame->seqno=%lu, frame->payload.type=%lu frame->len=%lu",
            iacpPeerIdent(iacp), fid, frame->seqno, frame->payload.type, frame->payload.len
        );
    }
    if (iacpGetDebug(iacp) > 2) {
        logioHexDump(iacp->lp, LOG_INFO, frame->payload.data, frame->payload.len);
    }
    if (iacpGetDebug(iacp) > 1) {
        logioMsg(iacp->lp, LOG_INFO, "%s: %s: frame->auth.id=0x%08x, frame->auth.len=%d",
            iacpPeerIdent(iacp), fid, frame->auth.id, frame->auth.len
        );
    }
    if (iacpGetDebug(iacp) > 2) {
        if (frame->auth.len > 0) logioHexDump(iacp->lp, LOG_INFO, frame->auth.sig, frame->auth.len);
    }

    return TRUE;
}

/* Revision History
 *
 * $Log: send.c,v $
 * Revision 1.7  2004/01/29 18:45:50  dechavez
 * cleaned up logic in read/write loops, log ignored errors
 *
 * Revision 1.6  2003/12/04 23:21:50  dechavez
 * added EWOULDBLOCK to list of ignored select failures
 *
 * Revision 1.5  2003/11/19 23:38:31  dechavez
 * tuned debug levels, added some casts to calm compiler
 *
 * Revision 1.4  2003/11/04 20:03:15  dechavez
 * adjust timeout setup to account for switch from sec to msec
 *
 * Revision 1.3  2003/10/16 16:36:31  dechavez
 * Many bug fixes and enhancements, to numerous to mention.
 *
 * Revision 1.2  2003/06/30 18:47:52  dechavez
 * log error messages instead of error codes
 *
 * Revision 1.1  2003/06/09 23:50:27  dechavez
 * initial release
 *
 */
