#!/usr/bin/env python

# Much of the following is borrowed from SciPy (www.scipy.org) in order to
# decouple it from SciPy itself (sorry).

import Numeric
import string # XXX
from Numeric import Float,concatenate,absolute,ones,zeros,resize,allclose,arange,asarray,array,tan,sqrt,exp,where,pi,sinh,cosh

def _poly(roots): # I apologize for this very crude hack!
    pol = array([1.])
    for x in roots:
        pol = concatenate([pol,[0]]) + concatenate([[0], -pol*x])
    return pol

def real(val): # from scipy
    aval = asarray(val)
    if aval.typecode() in ['F', 'D']:
        return aval.real
    else:
        return aval

def _comb(N,k): # from scipy.comb(), but MODIFIED!
    if (k > N) or (N < 0) or (k < 0):
        return 0L
    N,k = map(long,(N,k))
    top = N
    val = 1L
    while (top > (N-k)):
        val *= top
        top -= 1
    n = 1L
    while (n < k+1L):
        val /= n
        n += 1
    return val

def atleast_1d(*arys): # from scipy
    """ Force a sequence of arrays to each be at least 1D.

         Description:
            Force an array to be at least 1D.  If an array is 0D, the 
            array is converted to a single row of values.  Otherwise,
            the array is unaltered.
         Arguments:
            *arys -- arrays to be converted to 1 or more dimensional array.
         Returns:
            input array converted to at least 1D array.
    """
    res = []
    for ary in arys:
        ary = asarray(ary)
        if len(ary.shape) == 0:
            result = array([ary[0]])
        else:
            result = ary
        res.append(result)
    if len(res) == 1:
        return res[0]
    else:
        return res

def buttap(N): # from scipy
    """Return (z,p,k) zero, pole, gain for analog prototype of an Nth
    order Butterworth filter."""
    z = []
    n = arange(1,N+1)
    p = exp(1j*(2*n-1)/(2.0*N)*pi)*1j
    k = 1
    return z, p, k

def zpk2tf(z,p,k): # from scipy
    """Return polynomial transfer function representation from zeros
    and poles

    Inputs:

      z, p --- sequences representing the zeros and poles.
      k --- system gain.

    Outputs: (b,a)

      b, a --- numerator and denominator polynomials.
    """
    z = atleast_1d(z)
    k = atleast_1d(k)
    if len(z.shape) > 1:
        temp = _poly(z[0])
        b = zeros((z.shape[0], z.shape[1]+1), temp.typecode())
        if len(k) == 1:
            k = [k[0]]*z.shape[0]
        for i in range(z.shape[0]):
            b[i] = k[i] * _poly(z[i])
    else:
        b = k * _poly(z)
    a = _poly(p)
    return b, a

def normalize(b,a): # from scipy
    """Normalize polynomial representation of a transfer function.
    """
    b,a = map(atleast_1d,(b,a))
    if len(a.shape) != 1:
        raise ValueError, "Denominator polynomial must be rank-1 array."
    if len(b.shape) > 2:
        raise ValueError, "Numerator polynomial must be rank-1 or rank-2 array."
    if len(b.shape) == 1:
        b = asarray([b],b.typecode())
    while a[0] == 0.0 and len(a) > 1:
        a = a[1:]
    while allclose(b[:,0], 0, rtol=1e-14) and (b.shape[-1] > 1):
        b = b[:,1:]
    if b.shape[0] == 1:
        b = b[0]
    outb = b * (1.0) / a[0]
    outa = a * (1.0) / a[0]
    return outb, outa

def lp2lp(b,a,wo=1.0): # from scipy
    """Return a low-pass filter with cuttoff frequency wo
    from a low-pass filter prototype with unity cutoff frequency.
    """
    a,b = map(atleast_1d,(a,b))
    if type(wo) is type(a):
        wo = wo[0]
    wo = float(wo)
    d = len(a)
    n = len(b)
    M = max((d,n))
    pwo = pow(wo, arange(M-1,-1,-1))
    start1 = max((n-d,0))
    start2 = max((d-n,0))
    b = b * pwo[start1]/pwo[start2:]
    a = a * pwo[start1]/pwo[start1:]
    return normalize(b, a)

def lp2hp(b,a,wo=1.0): # from scipy
    """Return a high-pass filter with cuttoff frequency wo
    from a low-pass filter prototype with unity cutoff frequency.
    """
    a,b = map(atleast_1d,(a,b))
    if type(wo) is type(a):
        wo = wo[0]
    d = len(a)
    n = len(b)
    if wo != 1:
        pwo = pow(wo, arange(max((d,n))))
    else:
        pwo = ones(max((d,n)),b.typecode())
    if d >= n:
        outa = a[::-1] * pwo
        outb = resize(b,(d,))
        outb[n:] = 0.0
        outb[:n] = b[::-1] * pwo[:n]
    else:
        outb = b[::-1] * pwo
        outa = resize(a,(n,))
        outa[d:] = 0.0
        outa[:d] = a[::-1] * pwo[:d]

    return normalize(outb, outa)

def lp2bp(b,a,wo=1.0, bw=1.0): # from scipy, modified
    """Return a band-pass filter with center frequency wo and bandwidth bw
    from a low-pass filter prototype with unity cutoff frequency.
    """
    a,b = map(atleast_1d,(a,b))
    D = len(a) - 1
    N = len(b) - 1
    artype = b.typecode()
    if artype not in ['F','D','f','d']:
        artype = Float
    ma = max([N,D])
    Np = N + ma
    Dp = D + ma

    bprime = zeros(Np+1, 'D')
    aprime = zeros(Dp+1, 'D')

    wosq = wo*wo
    for j in range(Np+1):
        val = 0.0
        for i in range(0,N+1):
            for k in range(0,i+1):
                if ma-i+2*k == j:
                    val += _comb(i,k)*b[N-i]*(wosq)**(i-k) / bw**i
        bprime[Np-j] = val
    for j in range(Dp+1):
        val = 0.0
        for i in range(0,D+1):
            for k in range(0,i+1):
                if ma-i+2*k == j:
                    val += _comb(i,k)*a[D-i]*(wosq)**(i-k) / bw**i
        aprime[Dp-j] = val

    bprime = bprime.astype(artype)
    aprime = aprime.astype(artype)

    return normalize(bprime, aprime)

def lp2bs(b,a,wo=1,bw=1): # from scipy, slightly modified
    """Return a band-stop filter with center frequency wo and bandwidth bw
    from a low-pass filter prototype with unity cutoff frequency.
    """
    a,b = map(atleast_1d,(a,b))
    D = len(a) - 1
    N = len(b) - 1
    artype = b.typecode()
    if artype not in ['F','D','f','d']:
        artype = Float
    M = max([N,D])
    Np = M + M
    Dp = M + M
    bprime = zeros(Np+1,artype)
    aprime = zeros(Dp+1,artype)
    wosq = wo*wo
    for j in range(Np+1):
        val = 0.0
        for i in range(0,N+1):
            for k in range(0,M-i+1):
                if i+2*k == j:
                    val += _comb(M-i,k)*b[N-i]*(wosq)**(M-i-k) * bw**i
        bprime[Np-j] = val
    for j in range(Dp+1):
        val = 0.0
        for i in range(0,D+1):
            for k in range(0,M-i+1):
                if i+2*k == j:
                    val += _comb(M-i,k)*a[D-i]*(wosq)**(M-i-k) * bw**i
        aprime[Dp-j] = val

    return normalize(bprime, aprime)


def bilinear(b,a,fs=1.0): # from scipy, slightly modified
    """Return a digital filter from an analog filter using the bilinear transform.

    The bilinear transform substitutes (z-1) / (z+1) for s
    """
    fs =float(fs)
    a,b = map(atleast_1d,(a,b))
    D = len(a) - 1
    N = len(b) - 1
    artype = Float
    M = max([N,D])
    Np = M
    Dp = M
    bprime = zeros(Np+1,artype)
    aprime = zeros(Dp+1,artype)
    for j in range(Np+1):
        val = 0.0
        for i in range(N+1):
             for k in range(i+1):
                for l in range(M-i+1):
                    if k+l == j:
                        val += _comb(i,k)*_comb(M-i,l)*b[N-i]*pow(2*fs,i)*(-1)**k
        bprime[j] = real(val)
    for j in range(Dp+1):
        val = 0.0
        for i in range(D+1):
            for k in range(i+1):
                for l in range(M-i+1):
                    if k+l == j:
                        val += _comb(i,k)*_comb(M-i,l)*a[D-i]*pow(2*fs,i)*(-1)**k
        aprime[j] = real(val)

    return normalize(bprime, aprime)


def IIR_CoefficientsButterworth(fsamp, order, wn, btype='bandpass'):

    wn = asarray(wn)
    fs = 2.0 # normalized sampling frequency (1. would be fNyquist)
    wn = wn*fs/fsamp
    warped = 2*fs*tan(pi*wn/fs)
    if btype in ['lowpass', 'highpass']:
        wo = warped
    else:
        bw = warped[1] - warped[0]
        wo = sqrt(warped[0]*warped[1])

    typefunc = buttap
    z, p, k = typefunc(order)
    print "z=", z
    print "p=", p
    print "k=", k

    b, a = zpk2tf(z,p,k)

    if btype == 'lowpass':
        b, a = lp2lp(b,a,wo=wo)
    elif btype == 'highpass':
        b, a = lp2hp(b,a,wo=wo)
    elif btype == 'bandpass':
        b, a = lp2bp(b,a,wo=wo,bw=bw)
    elif btype == 'bandpass':
#   else: # 'bandstop'
        b, a = lp2bs(b,a,wo=wo,bw=bw)
    else:
        raise ValueError, "illegal argument btype='%s'" % btype

    # Find discrete equivalent if necessary
    b, a = bilinear(b, a, fs=fs)

    return b,a

def iirfilter(N, Wn, rp=None, rs=None, btype='band', analog=0, ftype='butter', output='ba'):
    """IIR digital and analog filter design given order and critical points.
    
    Description:
    
      Design an Nth order lowpass digital or analog filter and return the filter
      coefficients in (B,A) (numerator, denominator) or (Z,P,K) form.          

    Inputs:

      N -- the order of the filter.
      Wn -- a scalar or length-2 sequence giving the critical frequencies.
      rp, rs -- For chebyshev and elliptic filters provides the maximum ripple
                in the passband and the minimum attenuation in the stop band.
      btype -- the type of filter (lowpass, highpass, bandpass, or bandstop).
      analog -- non-zero to return an analog filter, otherwise
                a digital filter is returned.
      ftype -- the type of IIR filter (Butterworth, Cauer (Elliptic),
               Bessel, Chebyshev1, Chebyshev2)
      output -- 'ba' for (b,a) output, 'zpk' for (z,p,k) output.
    
    SEE ALSO butterord, cheb1ord, cheb2ord, ellipord
    """

    ftype, btype, output = map(string.lower, (ftype, btype, output))
    Wn = asarray(Wn)
    try:
        btype = band_dict[btype]
    except KeyError:
        raise ValueError, "%s is an invalid bandtype for filter." % btype

    try:
        typefunc = filter_dict[ftype][0]
    except KeyError:
        raise ValueError, "%s is not a valid basic iir filter." % ftype


    if output not in ['ba', 'zpk']:
        raise ValueError, "%s is not a valid output form." % output

    #pre-warp frequencies for digital filter design
    if not analog:
        fs = 2.0
        warped = 2*fs*tan(pi*Wn/fs)
    else:
        warped = Wn

    # convert to low-pass prototype
    if btype in ['lowpass', 'highpass']:
        wo = warped
    else:
        bw = warped[1] - warped[0]
        wo = sqrt(warped[0]*warped[1])

    # Get analog lowpass prototype
    if typefunc in [buttap, besselap]:
        z, p, k = typefunc(N)
    elif typefunc == cheb1ap:
        if rp is None:
            raise ValueError, "passband ripple (rp) must be provided to design a Chebyshev I filter."
        z, p, k = typefunc(N, rp)
    elif typefunc == cheb2ap:
        if rs is None:
            raise ValueError, "stopband atteunatuion (rs) must be provided to design an Chebyshev II filter."
        z, p, k = typefunc(N, rs)
    else:  # Elliptic filters
        if rs is None or rp is None:
            raise ValueErrro, "Both rp and rs must be provided to design an elliptic filter."
        z, p, k = typefunc(N, rp, rs)

    b, a = zpk2tf(z,p,k)

    # transform to lowpass, bandpass, highpass, or bandstop
    if btype == 'lowpass':
        b, a = lp2lp(b,a,wo=wo)
    elif btype == 'highpass':
        b, a = lp2hp(b,a,wo=wo)
    elif btype == 'bandpass':
        b, a = lp2bp(b,a,wo=wo,bw=bw)
    else: # 'bandstop'
        b, a = lp2bs(b,a,wo=wo,bw=bw)



    # Find discrete equivalent if necessary
    if not analog:
        b, a = bilinear(b, a, fs=fs)

    # Transform to proper out type (pole-zero, state-space, numer-denom)    
    if output == 'zpk':
        return tf2zpk(b,a)
    else:
        return b,a


def butter(N, Wn, btype='low', analog=0, output='ba'):
    """Butterworth digital and analog filter design.

    Description:

      Design an Nth order lowpass digital or analog Butterworth filter
      and return the filter coefficients in (B,A) or (Z,P,K) form.

    See also buttord.
    """
    return iirfilter(N, Wn, btype=btype, analog=analog, output=output, ftype='butter')

def cheby1(N, rp, Wn, btype='low', analog=0, output='ba'):
    """Chebyshev type I digital and analog filter design.

    Description:

      Design an Nth order lowpass digital or analog Chebyshev type I filter
      and return the filter coefficients in (B,A) or (Z,P,K) form.

    See also cheb1ord.      
    """
    return iirfilter(N, Wn, rp=rp, btype=btype, analog=analog, output=output, ftype='cheby1')

def cheby2(N, rs, Wn, btype='low', analog=0, output='ba'):
    """Chebyshev type I digital and analog filter design.

    Description:

      Design an Nth order lowpass digital or analog Chebyshev type I filter
      and return the filter coefficients in (B,A) or (Z,P,K) form.

    See also cheb2ord.
    """
    return iirfilter(N, Wn, rs=rs, btype=btype, analog=analog, output=output, ftype='cheby2')

def ellip(N, rp, rs, Wn, btype='low', analog=0, output='ba'):
    """Elliptic (Cauer) digital and analog filter design.

    Description:

      Design an Nth order lowpass digital or analog elliptic filter
      and return the filter coefficients in (B,A) or (Z,P,K) form.

    See also ellipord.
    """
    return iirfilter(N, Wn, rs=rs, rp=rp, btype=btype, analog=analog, output=output, ftype='elliptic')

def bessel(N, Wn, btype='low', analog=0, output='ba'):
    """Bessel digital and analog filter design.

    Description:

      Design an Nth order lowpass digital or analog Bessel filter
      and return the filter coefficients in (B,A) or (Z,P,K) form.

    """
    return iirfilter(N, Wn, btype=btype, analog=analog, output=output, ftype='bessel')


def maxflat():
    pass

def yulewalk():
    pass

def cheb1ap(N,rp):
    """Return (z,p,k) zero, pole, gain for Nth order Chebyshev type I
    lowpass analog filter prototype with rp decibels of ripple
    in the passband.
    """
    z = []
    eps = sqrt(10**(0.1*rp)-1.0)
    n = arange(1,N+1)
    mu = 1.0/N * log((1.0+sqrt(1+eps*eps)) / eps)
    theta = pi/2.0 * (2*n-1.0)/N
    p = -sinh(mu)*Num.sin(theta) + 1j*cosh(mu)*cos(theta)
    k = MLab.prod(-p).real
    if N % 2 == 0:
        k = k / sqrt((1+eps*eps))
    return z, p, k
    pass

def cheb2ap(N,rs):
    """Return (z,p,k) zero, pole, gain for Nth order Chebyshev type II
    lowpass analog filter prototype with rs decibels of ripple
    in the stopband.
    """
    de = 1.0/sqrt(10**(0.1*rs)-1)
    mu = arcsinh(1.0/de)/N

    if N % 2:
        m = N - 1
        n = Num.concatenate((Num.arange(1,N-1,2),Num.arange(N+2,2*N,2)))
    else:
        m = N
        n = Num.arange(1,2*N,2)

    z = conjugate(1j / cos(n*pi/(2.0*N)))
    p = exp(1j*(pi*Num.arange(1,2*N,2)/(2.0*N) + pi/2.0))
    p = sinh(mu) * p.real + 1j*cosh(mu)*p.imag
    p = 1.0 / p
    k = (MLab.prod(-p)/MLab.prod(-z)).real
    return z, p, k




def besselap():
    pass


if __name__ == '__main__':
    b,a = IIR_CoefficientsButterworth(20., 2, [2.5,3])

    print a
    print b

filter_dict = {'butter': [buttap],
               'butterworth' : [buttap],
#              'cauer' : [ellipap],
#              'elliptic' : [ellipap],
#              'ellip' : [ellipap],
#              'bessel' : [besselap],
#              'cheby1' : [cheb1ap],
#              'chebyshev1' : [cheb1ap],
#              'chebyshevi' : [cheb1ap],
#              'cheby2' : [cheb2ap],
#              'chebyshev2' : [cheb2ap],
#              'chebyshevii' : [cheb2ap]
               }

band_dict = {'band':'bandpass',
             'bandpass':'bandpass',
             'pass' : 'bandpass',
             'bp':'bandpass',
             'bs':'bandstop',
             'bandstop':'bandstop',
             'bands' : 'bandstop',
             'stop' : 'bandstop',
             'l' : 'lowpass',
             'low': 'lowpass',
             'lowpass' : 'lowpass',
             'high' : 'highpass',
             'highpass' : 'highpass',
             'h' : 'highpass'
             }

