#!/usr/bin/env python

from    Numeric import cos, pi
import  os, glob
import  seis
import  seis._cmt
from seis import ttt, delta

cmtdir = os.getenv("CMTDIR")

def __dek_iter(dek):
    """
    Iterates over the events within a dek file
    """
    f = file(dek)
    while 1:
        try:
            evt = seis._cmt.read(f)
            yield evt
        except EOFError:
            break


def __read_dek(dek):
    """
    Reads one dek file and returns an event dict
    """
    d = {}

    for evt in __dek_iter(dek):
        d[evt.label] = evt

    return d


def read(begin=None, end=None):
    """
    Reads the CMT catalog

    Only events that fall within the specified begin and
    end times are considered.
    """

    if not begin:
        begin = seis.Time("1900-01-01 00:00:00")
    if not end:
        end   = seis.Time("2100-01-01 00:00:00")

    cmt = {}
    pat1 = "%04d-%02d.dek" % (begin.yr, begin.mo)
    pat2 = "%04d-%02d.dek" % (  end.yr,   end.mo)

    for dek in glob.glob(cmtdir + "/????/????-??.dek"):
        if pat1 <= dek.split("/")[-1] <= pat2:
            d = __read_dek(dek)
            # update events that fall within the time window
            for key in d:
                evt = d[key]
                if begin <= evt.time <= end:
                    cmt[key] = evt

    return cmt


def select(cmt, lat_range=None, lon_range=None, dep_range=None,
                mag_range=None, tim_range=None, del_range=None):

    if not lat_range: lat_range = ( -90.,  90.)
    if not lon_range: lon_range = (-180., 180.) 
    if not dep_range: dep_range = (   0., 800.)
    if not mag_range: mag_range = (   0.,  10.)
    if not tim_range: tim_range = seis.Time("1900-01-01 00:00:00"), \
                                  seis.Time("2100-01-01 00:00:00")
    lonmin, lonmax = lon_range
    latmin, latmax = lat_range
    depmin, depmax = dep_range
    magmin, magmax = mag_range
    tmin,   tmax   = tim_range

    if lonmax < lonmin:
        lonmax += 360
    clon = 0.5*(lonmax+lonmin)
    wlon = 0.5*(lonmax-lonmin)
    xcos = cos(wlon*pi/180)

    result = {}

    for label in cmt:
        evt = cmt[label]
        if not tmin <= evt.time <= tmax: continue
        if not latmin <= evt.lat <= latmax: continue
        if not cos((evt.lon-clon)*pi/180) >= xcos: continue
        if not depmin <= evt.dep <= depmax: continue
        if not magmin <= evt.mb <= magmax: continue

        result[label] = evt

    return result


pi180 = pi/180.

def _cmtradpat(azi, ih, str, dip, rak):
    pi180 = pi/180.
    ih    = pi180*ih
    phi   = pi180*(azi-str)
    dip   = pi180*dip
    lam   = pi180*rak

    rsv =     sin(lam) * cos(2*dip)* cos(2*ih) * sin(phi) \
        -     cos(lam) * cos(dip)  * cos(2*ih) * cos(phi) \
        + 0.5*cos(lam) * sin(dip)  * sin(2*ih) * sin(2*phi) \
        - 0.5*sin(lam) * sin(2*dip)* sin(2*ih) * (1+sin(phi)**2)

    rsh =     cos(lam) * cos(dip)  * cos(ih)   * sin(phi) \
        +     cos(lam) * sin(dip)  * sin(ih)   * cos(2*phi) \
        +     sin(lam) * cos(2*dip)* cos(ih)   * cos(phi) \
        - 0.5*sin(lam) * sin(2*dip)* sin(ih)   * sin(2*phi)

    polsvsh = (arctan2(rsh, rsv)/pi180 + 360 ) % 360
    polqt   = (arctan2(rsh,-rsv)/pi180 + 360 ) % 360

    return rsh, rsv, polsvsh, polqt

def cmtradpat(azi, ih, cmt):
    return _cmtradpat(azi, ih, cmt.str1, cmt.dip1, cmt.rak1)

def radpat(evt, sta, phase="S"):

    lat, lon, dep = evt.coord
    dist, azi, baz = delta.delazi2(evt, sta)
    arrlist = ttt.compute(dist, dep)
    arr = arrlist.findPhase(phase)
    ih = arr.toff

    return cmtradpat(azi, ih, evt)

