"""
ttt module for the computation of teleseismic traveltimes.

Functions:

    compute(delta, depth)

Classes:

    Arrival

    ArrivalList

Members of class ArrivalList:

    findPhase (phase)
    findFirstOf ([ phases ])
"""
import Numeric, os, sys
import seis._ttt

tt = seis._ttt.ttt(branch="basic")

#olddir = os.path.abspath(os.curdir)
#tttdir = "/".join(_ttt.__file__.split("/")[:-1]) + "/tables"
#os.chdir(tttdir)
#_ttt.ttinit()
#os.chdir(olddir)
#del olddir



def enumerate_reverse(xlist):
    return [(i-1,xlist[i-1]) for i in range(len(xlist), 0, -1)]

class Arrival:
    """
    Body wave arrival

    Arrival.phase:  phase code, e.g. "P", "pP", "PKiKP", etc.
    Arrival.time:   travel time in seconds
    Arrival.slow:   angular slowness (in seconds/degree)
    Arrival.dtdh:   derivative dt/dh (change of t.t. with source depth)
    Arrival.dddp:   derivative dd/dp
    Arrival.toff:   takeoff angle at source (degrees)
    """

    def __init__(s, phase="", time=0., slow=0., dtdh=0., dddp=0., toff=0.):
        s.phase = phase
        s.time, s.slow, s.dtdh, s.dddp, s.toff = \
            float(time), float(slow), float(dtdh), float(dddp), float(toff)

    def __repr__(s):
        return "%-8s %7.2f %6.3f %8.4f %8.4f %6.2f" % \
            (s.phase, s.time, s.slow, s.dtdh, s.dddp, s.toff)

    __str__ = __repr__

class ArrivalList(list):
    """
    List of body wave arrivals

    Consists of a list of 'Arrival' objects, and in addition to
    methods of the 'list' builtin type, provides the following
    methods:

    ArrivalList.findPhase(string):
    ArrivalList.findFirstOf(list):
    """
    def findPhase(self, phase):
        for a in self:
            if a.phase == phase:
                return a
        return None

    def findFirstOf(self, phlist):
        for a in self:
            if a.phase in phlist:
                return a
        return None

    def remove(self, excl_regex):
        import re
        if type(excl_regex)==type("string"):
            excl_regex = [ excl_regex ]
        elif type(excl_regex)!=type([]):
            raise ValueError, "don't know how to remove " + str(excl_regex)
        for e in excl_regex:
            regex = re.compile(e)
            for i,a in enumerate_reverse(self):
                if regex.search(a.phase):
                    del self[i]

    def keepOnly(self, keep_regex):
        import re
        if type(keep_regex)==type("string"):
            keep_regex = [ keep_regex ]
        elif type(keep_regex)!=type([]):
            raise ValueError, "don't know how to keep " + str(keep_regex)
        regex = re.compile('(' + '|'.join(keep_regex) + ')')
        for i,a in enumerate_reverse(self):
            if not regex.search(a.phase):
                del self[i]

    def filter(self, keep_regex):
        self.keepOnly(keep_regex)

    def __repr__(self):
        return '\n'.join([str(a) for a in self])

    def toDict(self):
        d = {}
        for a in self:
            d[a.phase] = a
        return d

def compute(delta, depth, excl=None):
    """
    Computes the travel times of the most important teleseismic
    set of body wave phases for a given source depth 'dep' and  
    epicentral distance 'del'.
    """
#   if depth<0.1:
#       sys.stderr.write("Warning: Changed too shallow source depth "\
#                        "of %g to 0.1 km\n" % depth)
#       depth = 0.1

    phcd, ttim, dtdd, dtdh, dddp, toff = tt.compute(delta, depth)

    arrlist = ArrivalList()
    for i in xrange(len(phcd)):
        arr = Arrival(phcd[i], ttim[i], dtdd[i], dtdh[i], dddp[i], toff[i])
        arrlist.append(arr)
    if excl:
        arrlist.remove(excl)

    return arrlist

if __name__ == '__main__':

    arrlist = compute(30., 0.)
    print arrlist

