#!/usr/bin/env python

import os, re, sys, time
import seis, seiscomp, seis.delta
from Numeric import sqrt, exp, average, array

from seiscomp.autoloc_location import Origin, LocationFailed, dist_wt

LocSAT_dir = seiscomp.root_dir + "/lib/locsat"
LocSAT_exe = LocSAT_dir + "/bin/LocSAT"
LocSAT_tab = LocSAT_dir + "/tables"


class Station(seis.Station):

    def __init__(self, *args, **kwargs):
        seis.Station.__init__(self, *args, **kwargs)


class PhaseArrival(object): 

    __slots__ = ["id", "net", "sta", "pha", "time", "res", "imp", "dist", "azi", "used"]

    def __init__(self, **kwargs):
        self.used = True
        for kw in kwargs:
            setattr(self, kw, kwargs[kw])

        
class LocSAT(object):
    """ Wrapper class for the LocSAT earthquake location program """

    def __init__(self, stations, tmpdir='/tmp', logger=None,
                       verbose=False):
        self.command  = LocSAT_exe + " -s sta -c con -d in -o out >/dev/null 2>&1"
        import tempfile
        self.tmpdir = tempfile.mkdtemp(prefix='LocSATx.', dir=tmpdir)
        self.set_stations(stations)
        self.fixed_depth = 10
        self.verbose     = verbose
        # these are only for statistics:
        self.acc_time = 0.
        self.acc_call = 0
        self.failed = {}
        self.logger = logger

    def __del__(self):
        # Since at program termination 'os' may have already been deleted:
        import os
 
        for f in os.listdir(self.tmpdir):
            os.remove(self.tmpdir+'/'+f)
        os.rmdir(self.tmpdir)
        self.logger.write("LocSAT: Total number of calls: %d\n" \
                          "LocSAT: Total time spent: %10.3f\n" \
                           % (self.acc_call, self.acc_time))
        if self.acc_call==0: return
        self.logger.write("LocSAT: Avrg. time spent: %10.3f\n" \
                           % (self.acc_time/self.acc_call))

    def __call__(self, arrs=None, start=None,
                       autorefine=False, fixed_depth=False):
        return self.locate(arrs, start, autorefine, fixed_depth)

    def __add_hash(self, arrs):
        hash = self.__make_hash(arrs)
        self.failed[hash] = True

    def __make_hash(self, arrs):
        hash = [ a.id for a in arrs ]
        hash.sort()
        return " ".join(hash)

    def __test_hash(self, arrs):
        hash = self.__make_hash(arrs)
        if hash in self.failed:
            raise LocationFailed

    def locate2(self, org, autorefine=0, fixed_depth=False):
        # org is the input origin
        start = org.time, org.lat, org.lon, org.dep
        _org, _arr = self.locate(org.trgs, start, autorefine, fixed_depth)
        _org.trgs = org.trgs
        _org.arr  = _arr
        return _org

    def locate(self, arrs=None, start=None, autorefine=0, fixed_depth=None):
        """ Invokes LocSAT

        The arguments are a list of arrivals and optionally a
        start solution. If no start solution is supplied, LocSAT
        tries to find one itself. The list of arrivals must
        consist of objects with the following attributes: 'sta',
        'pha' and 'time', being the station code, phase code and
        relative arrival time (float value), respectively.

        Returns a tuple (org,arr) consisting of the origin (or
        None) and the (possibly modified) list of arrivals.

        If a valid origin is found, the list of arrivals is
        modified in such a way that e.g. residuals are added to
        each arrival after the computation. """

#       self.__test_hash(arrs)

        __t1 = time.time()

        st_time, st_lat, st_lon, st_dep = start

#        if fixed_depth:
#            if len(arrs) >= 15 and st_dep > 40: # XXX
#                    self.fixed_depth = None
#            else:   self.fixed_depth = st_dep
#        else:
#           self.fixed_depth = None

#       self.set_fixed_depth(st_dep) # XXX

        # sort arrivals according to starting time
        arrs = [ (a.time, a) for a in arrs if a.used ]  # decorate
        arrs.sort()                                     # sort
        arrs = [ a for t, a in arrs ]                   # undecorate
        first_arr = arrs[0]

        if self.verbose:
            msg = "Running LocSAT\n" \
                  "Start: %s %.2f %.2f\n" % (st_time, st_lat, st_lon)
            for a in arrs:
                msg += "%s %s\n" % (str(a.time), a.sta)
            self.logger.write(msg)

        cwd = os.getcwd()
        os.chdir(self.tmpdir)

        tmp = file("in", "w")
#       if type(self.fixed_depth) in (int,float):
#               depth = float(self.fixed_depth)
#       else:   depth = st_dep
        depth = st_dep # XXX

        strtime = st_time.strftime("%y%m%d %H%M")
        nused = len([1 for a in arrs if a.used])
        tmp.write(" %s %06.3f %.3f %.3f %.2f 0.0 %d\n" \
            % (strtime, st_time.sec, st_lat, st_lon, depth, nused))
        for i,arr in enumerate(arrs):
            if not arr.used:
                continue
            tmp.write("%d %s %s t d %.3f 1.0\n" % \
                    (i+1, arr.sta, arr.pha, arr.time-st_time))
        tmp.close()
#       if self.verbose:
#           os.system("cat in")

        self.__write_control_file(start, fixed_depth)
        self.fixed_depth = fixed_depth # used by __parse_output
        os.system(self.command)

        try:
            org, _arrs = self.__parse_output("out")
        except LocationFailed:
#           self.__add_hash(arrs)
            raise LocationFailed, "Location failed"
#       if self.verbose:
#           os.system("cat out")
        os.chdir(cwd)

        # update arrs
        # XXX this is still a bit messy!!! XXX
        for _a in _arrs:
            i = _a.id-1
            a = arrs[i]
            assert a.sta == _a.sta
            _a.net  = a.net
            _a.time = a.time
            _a.id   = a.id

# XXX HIER FEHLER BEI DER O.T.-BERECHNUNG #
        org.time = first_arr.time + org.time
        org.rms_res = sqrt(average(array([a.res*dist_wt(a.dist) for a in _arrs])**2))
#       org.rms_res = sqrt(average(array([a.res*a.imp for a in _arrs])**2))

        self.acc_time += time.time() - __t1
        self.acc_call += 1

#       org.arr  = _arrs
#       org.trgs =  arrs
#       if autorefine and len(_arrs) > 10 and org.rms_res < 50:
        if autorefine and org.rms_res < 20:
            org, _arrs = self._check_residuals(org, _arrs,
                                autorefine=autorefine-1)
        return org, _arrs

    def _check_residuals(self, org, arrs, autorefine=0):

        worst_res = 0.
#       print "XXX _check_residuals autorefine=%s" % autorefine
        for arr in arrs:
            if not arr.used:
                continue
            if abs(arr.res) > worst_res:
                worst_res = abs(arr.res)
                worst_arr = arr

        if worst_res < 6 or len(arrs)<=6:
            # if the worst residual is OK or number of stations is too low,
            # return the unmodified org 
            return org, arrs

        _arrs = arrs[:]
        worst_arr.used = False
        start = (org.time, org.lat, org.lon, org.dep)
        return self(_arrs, start, autorefine=autorefine)


    def set_fixed_depth(self, fixed_depth=False):
        self.fixed_depth = fixed_depth 

    def set_stations(self, stations):
        self.stations = stations
        self.__write_station_file()

    def __write_station_file(self):

        f = file(self.tmpdir+"/sta", "w")
        for key in self.stations:
            s = self.stations[key]
            f.write("%6s %8.4f %9.4f %6.1f\n" % (s.code, s.lat, s.lon, s.alt))
        f.close()

    def __write_control_file(self, start=None, fixed_depth=None):

        f = file(self.tmpdir+"/con", "w")
        f.write("%s tab\n/locsat/scorr TT\n" % LocSAT_tab)
        if start:
                f.write(" y ")
        else:   f.write(" n ")
        if type(fixed_depth) in (int,float):
                f.write("y ")
        else:   f.write("n ")
        f.write("y 0.90 -1.00 2.40 8 10\n")
        f.close()

    def __parse_output(self, fname):

        rex_end = re.compile(".*Location ran for .* iterations ... Converged.*")

        f = open(fname)
        lines = f.readlines()
        f.close()

        if not lines:
            raise LocationFailed, "failed to locate event"

        regex = re.compile(".*Final location estimate.*")
        # search lines until that regex is found
        for i,line in enumerate(lines):
            if regex.match(line):
                break
        lines = lines[i:]

        # read solution and errors

        org = Origin()

        try:
            # read coordinates and relative org time
            x = lines[1].split()
            assert x[0] == "Latitude:"
            org.lat, org.lat_err = float(x[1]), float(x[5])
            if x[3] == 'S': org.lat = -org.lat
            x = lines[2].split()
            assert x[0] == "Longitude:"
            org.lon, org.lon_err = float(x[1]), float(x[5])
            if x[3] == 'W': org.lon = -org.lon
            x = lines[3].split()
            assert x[0] == "Depth:"
            org.dep, org.dep_err = float(x[1]), float(x[4])
            x = lines[4].replace(':',' ').split() # XXX otherwise it can be like 'O.T.:-1126.339' !!!
            assert x[0] == "Relative"
            org.time, org.time_err = float(x[2]), float(x[5])

            # read confidence region
            x = lines[8].split()
            assert x[0] == "Semi-major"
            org.confreg_semmaj = float(x[2])
            x = lines[9].split()
            assert x[0] == "Semi-minor"
            org.confreg_semmin = float(x[2])
            x = lines[10].split()
            assert x[0] == "Major-axis"
            org.confreg_majaxi = float(x[2])
            if type(self.fixed_depth) in (int,float):
                    x = lines[11].split()
            else:   x = lines[12].split()
            assert x[0] == "Orig."
            org.confreg_timerr = float(x[3])
        except (ValueError, AssertionError, IndexError):
            raise LocationFailed, "failed to locate event"

        # read list of arrivals with residuals, importances etc.

        regex = re.compile("^ Ariv ID Statn  Phase  Type at *True Normalized.*")
        # search lines until that regex is found
        for i,line in enumerate(lines):
            if regex.match(line):
                break
        lines = lines[i+2:]
        
        arr = []
        for line in lines:
            a = PhaseArrival()
            try:
                x = line.split()
                a.sta, a.pha = x[1], x[2]
                a.dist,a.azi,baz = seis.delta.delazi2(org, self.stations[a.sta])
                a.id = int(x[0])
                a.res  = float(x[5])
                a.imp  = float(x[9])
            except (ValueError, IndexError):
                break

            arr.append(a)

        return org, arr
