import os, sys
from seiscomp._mseed import Record, Exception

# XXX TODO
#
# - improve error logging in iterator class
# - implement buffering to file in ring buffer class
# - use new yield keyword in iterator class (low priority)

class Input(object):
    """
    MSeed iterator class

    Iterates over the records in an open file-like object until EOF
    is reached. Particularly useful in 'for'-loops:
    for record in mseed.Input(f):
        ...
    """
    def __init__(self, infile, try_decode=True, logfile=sys.stderr, ignoreErrors=True):
        if type(infile) == str:
            infile = file(infile)
        self.infile  = infile
        self.logfile = logfile
        self.try_decode = try_decode

    def __del__(self):
        self.infile.close()

    def __getitem__(self, index):
        try:
            while 1: # until a valid record could be read
                rec = Record(self.infile)
                if self.try_decode:
                    try:
                        xyz = rec.data # once force decoding of data (XXX could be handled more elegantly)
                    except Exception,e:
                        self.logfile.write("A mseed.Exception '%s' was ignored\n" % e)
                        continue
                if rec.dtype==0:
                    continue
                return rec

        except Exception, e:
            self.logfile.write("A mseed.Exception '%s' was ignored\n" % e)
            return None # XXX
            
        except EOFError:
            raise IndexError    # end of a sequence


def sortRecordsBySequenceNumbers(recordList): # DSU sort
    deco = [ (rec.seq_no, rec) for rec in recordList ]
    deco.sort()
    return [ rec for seq_no, rec in deco ]

def sortRecordsByTime(recordList): # DSU sort
    deco = [ (rec.time, rec) for rec in recordList ]
    deco.sort()
    return [ rec for tim, rec in deco ]

def checkContinuity(recordList):
    seqNumbers = [ rec.seq_no for rec in recordList ]
    seqNumbers.sort() 
    try:
        assert seqNumbers == range(seqNumbers[0], seqNumbers[-1]+1)
    except:
        raise Exception, "There were inconsistencies in record sequence"

def gap_between(rec1, rec2, dtmax=0.5):
    """
    Determines if there is a gap between records rec1 and rec2.
    A maximum allowable time gap dtmax (in samples!) may be specified.
    """
    dt = abs(rec2.time-rec1.endtime)*rec1.fsamp
    if dt>dtmax:
        return True

    return False


def scan(ms, dtmax=1.): 

    if not isinstance(ms, list):
        ms = [ ms ]

    windows = [] 
    t1 = None
    for msfile in ms:
        for rec in Input(msfile):
            if t1 is None:
                t1, t2 = rec.begtime, rec.endtime
                continue 

            dt = (rec.begtime-t2)*rec.fsamp
            if abs(dt)>dtmax:
                windows.append((t1, t2))
                t1 = rec.begtime

            t2 = rec.endtime

    windows.append((t1, t2))
    return windows


class RingBuffer(object):
    """
    Simple ring buffer - experimental, don't use!
    """
    def __init__(self, nrec=256, bufdir=None, logfile=sys.stderr, verbosity=0):
        self.nrec      = nrec
        self.bufdir    = bufdir
        self.logfile   = logfile
        self.verbosity = verbosity

        if bufdir:
            self.__init_from_file(bufdir)

        self.reset()

    def __init_from_file(self, bufdir):
        # XXX
        if os.path.exists(bufdir):
            input = Input(fbuf)
            self.mbuf = sortRecordsByTime( [ r for r in input ] )
            del input

            if len(self.mbuf) > self.nrec:
                self.mbuf = self.mbuf[-self.nrec:]

            self.ptr = len(self.mbuf)

    def __del__(self):
        pass

    def __getitem__(self, index):
        if not self.mbuf: # empty memory buffer?
            raise IndexError
        n = len(self.mbuf)
        if n==1 and index!=0:
            raise IndexError

        n = (self.ptr + index - 1) % self.nrec
        return self.mbuf[n]

    def __add__(self, rec):
        self.append(rec)
        return self

    def __radd__(self, rec):
        self.append(rec)

    def append(self, rec):
        """
        Append a record to the ring buffer.
        """
        if self.mbuf:
            last = (self.ptr - 1) % self.nrec

            if self.mbuf[last].time == rec.time:
                return # avoid appending the same record twice

            if gap_between(self.mbuf[last], rec):
                ngap = (rec.time-self.mbuf[last].endtime)*rec.fsamp
                if self.logfile:
                    self.logfile.write("%s: ringbuffer: gap of %d samples " \
                        "starting %s - reset\n" \
                        % (rec.stream_id, ngap, str(self.mbuf[last].endtime)))
                self.reset()

        if len(self.mbuf) < self.nrec:
                self.mbuf.append(rec)
        else:   self.mbuf[self.ptr] = rec

        self.ptr = (self.ptr + 1) % self.nrec

    def reset(self):
        """
        Empties the buffer. Useful e.g. after a time gap was detected.
        """
        self.mbuf = []
        self.ptr  = 0

    def get(self):
        """
        Returns the whole ring buffer in proper time order.
        """
        return self.mbuf[self.ptr:] + self.mbuf[:self.ptr]

    def getTimeWindow(self, start, length):
        """
        Returns the specified time window from ring buffer as a list of
        records (but *not* as one contiguous chunk), which may contain
        less data than requested or may even be empty.
        """
        end = start+length

        return [ r for r in self.get()
                 if (r.time+r.nsamp/r.fsamp)>=start and r.begtime<=end ]


def rt_simul(f, speed=1.):
    """
    Iterator to simulate "real-time" MSeed input

    At startup, the first MSeed record is read. The following records are
    read in pseudo-real-time relative to the time of the first record,
    resulting in data flowing at realistic speed. This is useful e.g. for
    demonstrating real-time processing using real data of past events.

    The data in the input file may be multiplexed, but *must* be sorted by
    time, e.g. using 'mssort'.
    """
    import time

    rtime = time.time()
    stime = None
    for rec in Input(f):
        if stime is None:
            stime = rec.time
        tmax = stime + speed*(time.time() - rtime)
        if rec.time > tmax:
            time.sleep((rec.time-tmax+0.001)/speed)
        yield rec





#######################################################################
#### OBSOLETE STUFF FOLLOWS


# for backward compatibility only... to be removed soon.

class MSeedRecord(Record):

    def __init__(self, obj=None):
        sys.stderr.write("Please use mseed.Record instead of mseed.MSeedRecord\n")
        Record.__init__(self, obj) 


class MSeedInput(Input):

    def __init__(self, infile, ignoreErrors=True, logfile=sys.stderr):
        sys.stderr.write("Please use mseed.Input instead of mseed.MSeedInput\n")
        Input.__init__(self, infile, ignoreErrors, logfile) 

class MSeedException(Exception):

    def __init__(self, msg, exception=None):
        sys.stderr.write("Please use mseed.Exception instead of mseed.MSeedException\n")
        Exception.__init__(self, msg)

class MSeedRingBuffer(RingBuffer):

    def __init__(self, nrec=256, bufdir=None):
        sys.stderr.write("Please use mseed.RingBuffer instead of mseed.MSeedRingBuffer\n")
        RingBuffer.__init__(self, nrec, bufdir)

